import TurndownService, { Options } from "turndown";
import RSSParser from "rss-parser";
import { db } from "../db";
import interval from "interval-promise";
import client from "../client";
import sha1 from "sha1";
import html from "node-html-parser";
import request from "request-promise-native";
import { ICommand } from "./command";
import { Response } from "request";
import { TextChannel, Message, ReactionCollector, MessageReaction, Collector, User, Channel } from "discord.js";
import { Dictionary, ObjectChain } from "lodash";

const PREVIEW_CHAR_LIMIT = 300;
const verifyChannelID = db.get("newsPostVerifyChannel").value();

const reactionCollectors : Dictionary<ReactionCollector> = {};
const verifyMessageIdToPostId : Dictionary<string> = {};

const turndown = new TurndownService();
turndown.addRule("image", {
    filter: "img",
    replacement: () => ""
});
turndown.addRule("link", {
    filter: (node : HTMLElement, opts: Options) => node.nodeName === "A" && node.getAttribute("href") != null,
    replacement: (content: string, node: HTMLElement) => node.getAttribute("href")
});

const parser = new RSSParser();
const RSS_UPDATE_INTERVAL_MIN = 5;

function getThreadId(url: string) {
    let result = url.substring(url.lastIndexOf(".") + 1);
    if(result.endsWith("/"))
        result = result.substring(0, result.length - 1);
    return result;
}

interface PostItem {
    id: string,
    title: string,
    link: string,
    creator: string,
    contents: string,
    hash: string,
    messageId?: string,
    verifyMessageId?: string,
    type?: string
}

async function checkFeeds() {
    console.log(`Checking feeds on ${new Date().toISOString()}`);
    let feeds = db.get("rssFeeds").value();
    let oldNews = db.get("postedNewsGuids") as ObjectChain<any>;
    for(let feedEntry of feeds) {
        let feed = await parser.parseURL(feedEntry.url);
        if(feed.items.length == 0)
            continue;
        let printableItems = feed.items.sort((a : any, b: any) => a.isoDate.localeCompare(b.isoDate));
        if(printableItems.length > 0) {
            for(let item of printableItems) {
                let itemID = getThreadId(item.guid);
                let contents = null;
                
                try {
                    let res = await request(item.link, {resolveWithFullResponse: true}) as Response;
                    if(res.statusCode != 200) {
                        console.log(`Post ${itemID} could not be loaded because request returned status ${res.statusCode}`);
                        continue;
                    }

                    let rootNode = html.parse(res.body, {
                        pre: true,
                        script: false,
                        style: false
                    });

                    if(!(rootNode instanceof HTMLElement))
                        continue;

                    let opDiv = rootNode.querySelector("div.bbWrapper");

                    if (!opDiv) {
                        console.log(`No posts found for ${itemID}!`);
                        continue;
                    }
                    
                    contents = markdownify(opDiv.outerHTML, item.link);
                } catch(err){
                    console.log(`Failed to get html for item ${itemID} because ${err}`);
                    continue;
                }

                let itemObj = {
                    id: itemID,
                    title: item.title,
                    link: item.link,
                    creator: item.creator,
                    contents: `${contents.substr(0, Math.min(contents.length, PREVIEW_CHAR_LIMIT))}...`,
                    hash: sha1(contents),
                    messageId: null,
                    verifyMessageId: null,
                    type: null
                } as PostItem;

                if(oldNews.has(itemObj.id).value()){
                    let data = oldNews.get(itemObj.id).value();
                    // Old type, don't care
                    if(data === true)
                        continue;

                    // Add message ID to mark for edit
                    if(data.hash != itemObj.hash)
                        itemObj.messageId = data.messageId;
                    else 
                        continue;
                }

                if(!shouldVerify())
                    await sendNews(itemObj);
                else
                    await addVerifyMessage(itemObj);
            }
            let lastUpdateDate = printableItems[printableItems.length - 1].isoDate;
            console.log(`Setting last update marker on ${feedEntry.url} to ${lastUpdateDate}`);
            let rssFeeds = db.get("rssFeeds") as ObjectChain<any>;
            let entry = rssFeeds.find({ url: feedEntry.url}) as ObjectChain<any>;
            entry.assign({lastUpdate: lastUpdateDate}).write();
        }
    }
}

function initPendingReactors() {
    let verifyChannel = client.channels.get(verifyChannelID);
    db.get("newsCache").forOwn(async (i : any) => {
        let m = await tryFetchMessage(verifyChannel, i.verifyMessageId);
        let collector = m.createReactionCollector(isVerifyReaction, { maxEmojis: 1 });
        collector.on("collect", collectReaction)
        reactionCollectors[m.id] = collector;
        verifyMessageIdToPostId[m.id] = i.id;
    }).value();
}

async function addVerifyMessage(item : PostItem) {
    let verifyChannel = client.channels.get(verifyChannelID) as TextChannel;
    let cache = db.get("newsCache") as ObjectChain<any>;
    let oldNews = db.get("postedNewsGuids");
    let postedNews = db.get("postedNewsGuids");
    item.type = "🆕 ADD";

    if(postedNews.has(item.id).value())
        item.type = "✏️ EDIT";

    if(cache.has(item.id).value()) {
        let oldItem = cache.get(item.id).value();
        let oldMessage = await tryFetchMessage(verifyChannel, oldItem.verifyMessageId);
        if(oldMessage)
            await oldMessage.delete();
    }

    let newMessage = await verifyChannel.send(toVerifyString(item)) as Message;
    
    await newMessage.react("✅");
    await newMessage.react("❌");

    let collector = newMessage.createReactionCollector(isVerifyReaction, { maxEmojis: 1 });
    collector.on("collect", collectReaction)
    reactionCollectors[newMessage.id] = collector;
    verifyMessageIdToPostId[newMessage.id] = item.id;
    item.verifyMessageId = newMessage.id;
    cache.set(item.id, item);

    oldNews.set(item.id, {
        hash: item.hash,
        messageId: null
    });
    db.write();
}

function collectReaction(reaction : MessageReaction, collector: Collector<string, MessageReaction>) {
    let cache = db.get("newsCache") as ObjectChain<any>;
    let m = reaction.message;
    collector.stop();
    delete reactionCollectors[m.id];
    let postId = verifyMessageIdToPostId[m.id];
    let item = cache.get(postId).value() as PostItem;
    cache.unset(postId).write();
    m.delete();

    if(reaction.emoji.name == "✅")
        sendNews(item);
}

async function sendNews(item : PostItem) {
    let outChannel = db.get("feedOutputChannel").value();
    let oldNews = db.get("postedNewsGuids");

    let sentMessage = await postNewsItem(outChannel, item);
    oldNews.set(item.id, {
                    hash: item.hash,
                    messageId: sentMessage.id
                });
    db.write();
}

function isVerifyReaction(reaction : MessageReaction, user: User) {
    return (reaction.emoji.name == "✅" || reaction.emoji.name == "❌") && !user.bot;
}

async function tryFetchMessage(channel: Channel, messageId: string)  {
    try {
        if(!(channel instanceof TextChannel))
            return null;
        return await channel.fetchMessage(messageId);
    }catch(error){
        return null;
    }
}

function shouldVerify() {
    return verifyChannelID != "";
}

async function postNewsItem(channel: string, item: PostItem) : Promise<Message | null> {
    let newsMessage = toNewsString(item);
    let ch = client.channels.get(channel);

    if(!(ch instanceof TextChannel))
        return null;

    if(item.messageId) {
        let message = await tryFetchMessage(ch, item.messageId);
        if(message)
            return await message.edit(newsMessage);
        else 
            return await ch.send(newsMessage) as Message;
    } 
    else 
        return await ch.send(newsMessage) as Message;
}

function markdownify(htmStr: string, link: string) {
    return turndown.turndown(htmStr).replace(/( {2}\n|\n\n){2,}/gm, "\n").replace(link, "");
}

function toNewsString(item: PostItem) {
    return `**${item.title}**
Posted by ${item.creator}
${item.link}

${item.contents}`;
}

function toVerifyString(item: PostItem) {
    return `[${item.type}]
Post ID: **${item.id}**
    
${toNewsString(item)}
            
React with ✅ (approve) or ❌ (deny).`;
}

export default {
    commands: [
        {
            pattern: /^edit (\d+)\s+((.*[\n\r]*)+)$/i,
            action: async (msg, s, match) => {
                if(msg.channel.id != verifyChannelID)
                    return;
    
                let id = match[1];
                let newContents = match[2].trim();
                
                if(!db.get("newsCache").has(id).value()) {
                    msg.channel.send(`${msg.author.toString()} No unapproved news items with id ${id}!`);
                    return;
                }
                
                let newsCache = db.get("newsCache") as ObjectChain<any>;
                let item = newsCache.get(id).value() as PostItem;
    
                let editMsg = await tryFetchMessage(client.channels.get(verifyChannelID), item.verifyMessageId);
    
                if(!editMsg){
                    msg.channel.send(`${msg.author.toString()} No verify messafe found for ${id}! This is a bug: report to horse.`);
                    return;
                }
    
                item.contents = newContents;
                
                db.get("newsCache").set(id, item);
                db.write();
                await editMsg.edit(toVerifyString(item));
                await msg.delete();
            }
        }
    ],
    onStart: () => {
        initPendingReactors();
        interval(checkFeeds, RSS_UPDATE_INTERVAL_MIN * 60 * 1000);
    }
} as ICommand;